#include <iostream>
#include <cmath>

// Рассказываем компилятору, какие классы у нас есть
class Point;
class Figure;
class Circle;
class Rectangle;
class UnitedFigure;
class ComplementedFigure;
class IntersectedFigure;

// Описываем наши классы
class Point {
public:
    double x;
    double y;

    Point() = default;
    Point(double _x, double _y) : x(_x), y(_y) {}

    Point operator + (const Point& p) const {
        return Point {x + p.x, y + p.y};
    }
    Point operator - (const Point& p) const {
        return Point {x - p.x, y - p.y};
    }

    static Point max (const Point& p1, const Point& p2) {
        return Point {std::max(p1.x, p2.x), std::max(p1.y, p2.y)};
    }

    double vec_length () const {
        return sqrt(x*x + y*y);
    }
};

class Figure {
public:
    virtual double distance_to (const Point &p) const = 0;
    friend UnitedFigure operator + (Figure & f1, Figure & f2); // friend значит, что описание метода будет ниже
    friend ComplementedFigure operator - (Figure & f1, Figure & f2);
    friend IntersectedFigure operator & (Figure & f1, Figure & f2);
    bool is_point_into(const Point &p) const {
        return distance_to(p) <= 0;
    }
};

class Circle : public Figure {
    Point o;
    double r;

public:
    Circle (Point p, double _r) : o(p), r(_r) {}

    double distance_to (const Point &p) const override {
        return (o - p).vec_length() - r;
    }
};

class Rectangle : public Figure {
    Point a;
    Point b;

public:
    Rectangle (Point p1, Point p2) : a(p1), b(p2) {}

    double distance_to (const Point &p) const override {
        auto d = Point::max(a - p, p - b);
        return Point::max(d, Point {0, 0}).vec_length() + std::min(0.0, std::max(d.x, d.y));
    }
};

class UnitedFigure : public Figure {
    Figure &f1;
    Figure &f2;

public:
    UnitedFigure (Figure &_f1, Figure &_f2) : f1(_f1), f2(_f2) {}

    double distance_to(const Point &p) const override {
        return std::min(f1.distance_to(p), f2.distance_to(p));
    }
};

class ComplementedFigure : public Figure {
    Figure &f1;
    Figure &f2;

public:
    ComplementedFigure (Figure &_f1, Figure &_f2) : f1(_f1), f2(_f2) {}

    double distance_to(const Point &p) const override {
        return std::max(f1.distance_to(p), -f2.distance_to(p));
    }
};

class IntersectedFigure : public Figure {
    Figure &f1;
    Figure &f2;

public:
    IntersectedFigure (Figure &_f1, Figure &_f2) : f1(_f1), f2(_f2) {}

    double distance_to(const Point &p) const override {
        return std::max(f1.distance_to(p), f2.distance_to(p));
    }
};

// А вот и запазладые описания операторов
// Если писать их прямо в Figure, то компилятор будет ругаться на необъявленные классы UnitedFigure и прочие
UnitedFigure operator + (Figure & f1, Figure & f2) {
    return UnitedFigure{f1, f2};
}

ComplementedFigure operator - (Figure & f1, Figure & f2) {
    return ComplementedFigure{f1, f2};
}

IntersectedFigure operator & (Figure & f1, Figure & f2) {
    return IntersectedFigure{f1, f2};
}

void test_this_point (Figure& figure, Point M) {
    auto is_M_into_figure = figure.is_point_into(M);
    std::cout << "For point M(" << M.x << ";" << M.y << ") result is " << std::boolalpha << is_M_into_figure << std::endl;
}

// Инструментарий подготовлены
// Вот теперь пора творить!
int main() {
    // Создаём точку М и читаем её с клавиатуры
    Point M {};
    std::cout << "Enter a coordinates of the M(x, y) point: ";
    std::cin >> M.x >> M.y;

    // Создаём нужные примитивы
    Circle big(Point{0, 0}, 20);
    Circle small(Point{0, 0}, 10);
    Rectangle quarter_1(Point{0, 0}, Point{10, 10});
    Rectangle quarter_3(Point{0, 0}, Point{-10, -10});

    // Собираем из них нашу фигуру. Просто же? Я тоже так думаю
    auto quarter_1_and_3 = quarter_1 + quarter_3;
    auto inner_empty_space = small - quarter_1_and_3;
    auto figure = big - inner_empty_space;

    // Проверяем, находится ли точка М внутри
    auto is_M_into_figure = figure.is_point_into(M);

    // Выводим результат
    std::cout << std::boolalpha << is_M_into_figure << std::endl << std::endl;

    // Тестим заготовленные точки
    test_this_point(figure, Point{-3, 3});
    test_this_point(figure, Point{-10, 10});
    test_this_point(figure, Point{-15, 15});
    test_this_point(figure, Point{3, 3});
    test_this_point(figure, Point{10, 10});
    test_this_point(figure, Point{15, 15});

    return 0;
}
